'''Generate random plates.'''

import csv
import json
from pathlib import Path
import random
import sys

from .params import AssayParams, load_params


MODEL = 'Weyland-Yutani 470'
PLATE_HEIGHT = 4
PLATE_WIDTH = 4


def plates(options):
    '''Main driver.'''
    options.params = load_params(AssayParams, options.params)
    random.seed(options.params.seed)
    _create_files(options)


def _create_files(options):
    '''Create randomized plate files.'''
    for filename, sample_id, kind in _join_assay_data(options):
        _make_plate(
            options.params,
            sample_id,
            kind,
            Path(options.designs, filename),
            Path(options.readings, filename),
        )


def _generate(params, func):
    '''Make body of plate design or results.'''
    title_row = ['', *[chr(ord('A') + col) for col in range(PLATE_WIDTH)]]
    values = [
        [func(params, _make_placement) for col in range(PLATE_WIDTH)]
        for row in range(PLATE_HEIGHT)
    ]
    labeled = [[str(i + 1), *r] for (i, r) in enumerate(values)]
    return [title_row, *labeled]


def _join_assay_data(options):
    '''Get experiment type and plate filename from data.'''
    assays = json.load(open(options.assays, 'r'))
    experiments = {x['sample_id']: x['kind'] for x in assays['experiment']}
    plates = {p['filename']: p['sample_id'] for p in assays['plate']}
    return ((f, plates[f], experiments[plates[f]]) for f in plates)


def _make_head(kind, sample_id):
    '''Make head of plate.'''
    return [
        [MODEL, kind, sample_id],
        [],
    ]


def _make_placement(kind):
    '''Generate random placement of samples.'''
    placement = [[False for col in range(PLATE_WIDTH)] for row in range(PLATE_HEIGHT)]
    if kind == 'calibration':
        return placement, []
    columns = list(c for c in range(PLATE_WIDTH))
    random.shuffle(columns)
    columns = columns[:PLATE_HEIGHT]
    for r, row in enumerate(placement):
        row[columns[r]] = True
    return placement, columns


def _make_plate(params, sample_id, kind, design_file, readings_file):
    '''Generate an entire plate.'''
    placement, sample_locs = _make_placement(kind)

    design = [*_make_head('design', sample_id), *_generate(params, _make_treatment)]
    _save(design_file, _normalize_csv(design))

    readings = [*_make_head('readings', sample_id), *_generate(params, _make_reading)]
    _save(readings_file, _normalize_csv(readings))


def _make_reading(params, treated):
    '''Generate a single plate reading.'''
    mean = params.treated_val if treated else params.control_val
    value = max(0.0, random.gauss(mean, params.stdev))
    return f'{value:.02f}'


def _make_treatment(params, treated):
    '''Generate a single plate treatment.'''
    return params.treatment if treated else random.choice(params.controls)


def _normalize_csv(rows):
    required = max(len(r) for r in rows)
    for row in rows:
        row.extend([''] * (required - len(row)))
    return rows


def _save(filename, rows):
    '''Save as CSV.'''
    if not filename:
        csv.writer(sys.stdout).writerows(rows)
    else:
        csv.writer(open(filename, 'w'), lineterminator='\n').writerows(rows)



'''Parameter dataclasses and utilities.'''

from dataclasses import dataclass, field
from datetime import date, datetime
from importlib.resources import files
import json
from pathlib import Path
from typing import List


DATE_FORMAT = '%Y-%m-%d'
DEFAULT_START_DATE = datetime.strptime('2023-11-01', DATE_FORMAT)
DEFAULT_END_DATE = datetime.strptime('2023-11-10', DATE_FORMAT)
RADIUS = 200.0
PARAMETER_FILES = (
    'params/assays.json',
    'params/genomes.json',
    'params/grids.json',
    'params/samples.json',
    'params/sites.csv',
    'params/staff.json',
    'params/surveys.csv',
)


@dataclass
class AssayParams:
    '''Parameters for assay data generation.'''

    assay_staff: list
    assay_types: list
    assay_duration: list
    assay_plates: list
    control_val: float = 5.0
    controls: List[str] = field(default_factory=list)
    enddate: date = None
    experiments: int = 1
    filename_length: int = 8
    fraction: float = None
    invalid: float = 0.1
    seed: int = None
    startdate: date = None
    stdev: float = 3.0
    treated_val: float = 8.0
    treatment: str = None

    def __post_init__(self):
        '''Convert dates if provided.'''
        if self.startdate is None:
            self.startdate = DEFAULT_START_DATE
        else:
            self.startdate = datetime.strptime(self.startdate, DATE_FORMAT)

        if self.enddate is None:
            self.enddate = DEFAULT_END_DATE
        else:
            self.enddate = datetime.strptime(self.enddate, DATE_FORMAT)


@dataclass
class GenomeParams:
    '''Gene sequence parameters.'''
    snp_probs: list
    length: int
    num_genomes: int
    num_snp: int
    prob_other: float
    seed: int = None


@dataclass
class GridParams:
    '''Invasion percolation parameters.'''
    depth: int
    height: int
    seed: int
    width: int


@dataclass
class SampleParams:
    '''Sampled snail parameters.'''
    min_snail_size: float = None
    max_snail_size: float = None
    mutant: float = None
    normal: float = None
    seed: int = None


@dataclass
class StaffParams:
    '''Staff parameters.'''
    locale: str = 'hy_AM'
    num: int = None
    seed: int = None


def export_params(options):
    '''Export parameter files.'''
    outdir = Path(options.outdir)
    outdir.mkdir(parents=True, exist_ok=True)
    root = files(__name__.split('.')[0])
    for filename in PARAMETER_FILES:
        src = root.joinpath(filename)
        dst = outdir.joinpath(Path(filename).name)
        dst.write_bytes(src.read_bytes())


def load_params(cls, filename):
    '''Load parameters from file.'''
    return cls(**json.loads(Path(filename).read_text()))



'''Generate staff.'''

import faker
from pathlib import Path
import polars as pl
import random
import sys

from .params import StaffParams, load_params


def staff(options):
    '''Main driver.'''
    options.params = load_params(StaffParams, options.params)
    random.seed(options.params.seed)
    faker.Faker.seed(options.params.seed)
    fake = faker.Faker(options.params.locale)
    people = _make_people(options.params, fake)
    _save(options, people)


def _make_people(params, fake):
    '''Create people.'''
    people = [(i+1, fake.first_name(), fake.last_name()) for i in range(params.num)]
    return pl.DataFrame(people, schema=('staff_id', 'personal', 'family'), orient='row')


def _save(options, people):
    '''Save or show results.'''
    if options.outfile:
        people.write_csv(Path(options.outfile))
    else:
        people.write_csv(sys.stdout)



'''Generate grids using invasion percolation.'''

from .params import GridParams, load_params
from pathlib import Path
import polars as pl
import random


class Grid:
    '''Represent a lazy grid.'''

    def __init__(self, params):
        '''Record shared state.'''
        self._width = params.width
        self._height = params.height
        self._depth = params.depth
        self._grid = []
        for x in range(self._width):
            col = []
            for y in range(self._height):
                col.append(random.randint(1, self._depth))
            self._grid.append(col)
        self._candidates = {}

    def __getitem__(self, key):
        '''Get value at location.'''
        x, y = key
        return self._grid[x][y]

    def __setitem__(self, key, value):
        '''Set value at location.'''
        x, y = key
        self._grid[x][y] = value

    def __str__(self):
        '''Convert to printable string.'''
        rows = []
        for y in range(self.height - 1, -1, -1):
            rows.append(''.join('x' if self[x, y] == 0 else '.' for x in range(self.width)))
        return '\n'.join(rows)

    @property
    def depth(self):
        '''Get depth of grid.'''
        return self._depth

    @property
    def height(self):
        '''Get height of grid.'''
        return self._height

    @property
    def width(self):
        '''Get width of grid.'''
        return self._width

    def fill(self):
        '''Fill grid one cell at a time.'''
        x, y = self.width // 2, self.height // 2
        self[x, y] = - self[x, y]
        self.add_candidates(x, y)
        while True:
            x, y = self.choose_cell()
            self[x, y] = - self[x, y]
            if self.on_border(x, y):
                break

    def add_candidates(self, x, y):
        '''Add candidates around (x, y).'''
        for ix in (x - 1, x + 1):
            self.add_one_candidate(ix, y)
        for iy in (y - 1, y + 1):
            self.add_one_candidate(x, iy)

    def add_one_candidate(self, x, y):
        '''Add (x, y) if suitable.'''
        if (x < 0) or (x >= self.width) or (y < 0) or (y >= self.height):
            return
        if self[x, y] < 0:
            return

        value = self[x, y]
        if value not in self._candidates:
            self._candidates[value] = set()
        self._candidates[value].add((x, y))

    def adjacent(self, x, y):
        '''Is (x, y) adjacent to a filled cell?'''
        x_1, y_1 = x + 1, y + 1
        if (x > 0) and (self[x - 1, y] < 0):
            return True
        if (x_1 < self.width) and (self[x_1, y] < 0):
            return True
        if (y > 0) and (self[x, y - 1] < 0):
            return True
        if (y_1 < self.height) and (self[x, y_1] < 0):
            return True
        return False

    def choose_cell(self):
        '''Choose the next cell to fill.'''
        min_key = min(self._candidates.keys())
        available = list(sorted(self._candidates[min_key]))
        i = random.randrange(len(available))
        choice = available[i]
        del available[i]
        if not available:
            del self._candidates[min_key]
        else:
            self._candidates[min_key] = set(available)
        self.add_candidates(*choice)
        return choice

    def on_border(self, x, y):
        '''Is this cell on the border of the grid?'''
        if (x == 0) or (x == self.width - 1):
            return True
        if (y == 0) or (y == self.height - 1):
            return True
        return False


def grid(options):
    '''Main driver.'''
    options.params = load_params(GridParams, options.params)
    options.sites = pl.read_csv(options.sites)
    random.seed(options.params.seed)
    for row in options.sites.iter_rows(named=True):
        grid = Grid(options.params)
        grid.fill()
        _save(options.outdir, row['site_id'], grid)


def _save(outdir, site_id, grid):
    '''Save grid as CSV.'''
    Path(outdir).mkdir(parents=True, exist_ok=True)
    with open(Path(outdir, f'{site_id}.csv'), 'w') as writer:
        for y in range(grid.height - 1, -1, -1):
            values = (- grid[x, y] if grid[x, y] < 0 else 0 for x in range(grid.width))
            print(','.join((str(v) for v in values)), file=writer)



'''Generate assays.'''

from datetime import date, datetime, timedelta
import json
from pathlib import Path
import polars as pl
import random
import string

from .params import AssayParams, load_params


class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (date, datetime)):
            return obj.isoformat()


def assays(options):
    '''Main driver.'''
    assert options.params != options.outfile, 'Cannot use same filename for options and parameters'
    options.params = load_params(AssayParams, options.params)
    random.seed(options.params.seed)

    samples = _reload_samples(options)
    staff_ids = _reload_staff(options)
    result = _make_experiments(options.params, samples, staff_ids)
    _save(options.outfile, result)


def _reload_samples(options):
    '''Re-create sample genomic information.'''
    genomes = json.loads(Path(options.genomes).read_text())
    susc_loc = genomes['susceptible_loc']
    susc_base = genomes['susceptible_base']
    samples = pl.read_csv(options.samples)
    return [g[susc_loc] == susc_base for g in samples['sequence']]


def _reload_staff(options):
    '''Re-load staff information.'''
    return pl.read_csv(options.staff)['staff_id'].to_list()


def _make_experiments(params, samples, staff_ids):
    '''Create experiments and their data.'''
    kinds = list(params.assay_types)
    experiments = []
    performed = []
    plates = []

    num_samples = len(samples)
    keepers = set(random.sample(list(range(num_samples)), k=int(params.fraction * num_samples)))

    random_filename = _make_random_filename(params)
    for i, flag in enumerate(samples):
        if i not in keepers:
            continue

        sample_id = i + 1
        kind = random.choice(kinds)
        started, ended = _random_experiment_duration(params, kind)
        experiments.append(
            {'sample_id': sample_id, 'kind': kind, 'start': _round_date(started), 'end': _round_date(ended)}
        )

        num_staff = random.randint(*params.assay_staff)
        performed.extend(
            [{'staff_id': s, 'sample_id': sample_id} for s in random.sample(staff_ids, num_staff)]
        )

        if ended is not None:
            plates.extend(
                _random_plates(params, kind, sample_id, len(plates), started, random_filename)
            )

    invalidated = _invalidate_plates(params, staff_ids, plates)

    return {
        'experiment': experiments,
        'performed': performed,
        'plate': plates,
        'invalidated': invalidated
    }


def _invalidate_plates(params, staff_ids, plates):
    '''Invalidate a random set of plates.'''
    selected = [
        (i, p['date']) for (i, p) in enumerate(plates) if random.random() < params.invalid
    ]
    return [
        {
            'plate_id': plate_id,
            'staff_id': random.choice(staff_ids),
            'date': _random_date_interval(exp_date, params.enddate),
        }
        for (plate_id, exp_date) in selected
    ]


def _make_random_filename(params):
    '''Create a random filename generator.'''
    filenames = set([''])
    result = ''
    while True:
        while result in filenames:
            stem = ''.join(random.choices(string.hexdigits, k=params.filename_length)).lower()
            result = f'{stem}.csv'
        filenames.add(result)
        yield result


def _random_experiment_duration(params, kind):
    '''Choose random start date and end date for experiment.'''
    start = random.uniform(params.startdate.timestamp(), params.enddate.timestamp())
    start = datetime.fromtimestamp(start)
    duration = timedelta(days=random.randint(*params.assay_duration))
    end = start + duration
    end = None if end > params.enddate else end
    return start, end


def _random_plates(params, kind, sample_id, start_id, start_date, random_filename):
    '''Generate random plate data.'''
    return [
        {
            'plate_id': start_id + i + 1,
            'sample_id': sample_id,
            'date': _random_date_interval(start_date, params.enddate),
            'filename': next(random_filename),
        }
        for i in range(random.randint(*params.assay_plates))
    ]


def _random_date_interval(start_date, end_date):
    '''Choose a random end date (inclusive).'''
    if isinstance(start_date, date):
        start_date = datetime(*start_date.timetuple()[:3])
    choice = random.uniform(start_date.timestamp(), end_date.timestamp())
    choice = datetime.fromtimestamp(choice)
    return _round_date(choice)


def _round_date(raw):
    '''Round time to whole day.'''
    return None if raw is None else date(*raw.timetuple()[:3])


def _save(outfile, result):
    '''Save or show generated data.'''
    as_text = json.dumps(result, indent=4, cls=DateTimeEncoder)
    if outfile:
        Path(outfile).write_text(as_text)
    else:
        print(as_text)



'''Generate sample snails with genomes and locations.'''

import json
import math
import numpy as np
from pathlib import Path
import polars as pl
import random
from geopy.distance import lonlat, distance
import sys

from .params import SampleParams, load_params


LON_LAT_PRECISION = 5
READING_PRECISION = 1
SIZE_PRECISION = 1


def samples(options):
    '''Main driver.'''
    assert options.params != options.outfile, 'Cannot use same filename for options and parameters'
    options.params = load_params(SampleParams, options.params)
    options.surveys = pl.read_csv(options.surveys)
    options.sites = pl.read_csv(options.sites)
    random.seed(options.params.seed)

    genomes = json.loads(Path(options.genomes).read_text())
    grids = _load_grids(options)

    samples = _generate_samples(options, genomes, grids)
    _save(options, samples)


def _generate_samples(options, genomes, grids):
    '''Generate snail samples.

    Snails have mutant phenotype if they have the susceptible gene and
    are in a contaminated location.
    '''

    # Generate.
    samples = []
    for i, seq in enumerate(genomes['individuals']):
        survey_id, point, contaminated = _random_geo(options.sites, options.surveys, grids)
        limit = _size_limit(options, genomes, seq, contaminated)
        size = random.uniform(
            options.params.min_snail_size,
            options.params.min_snail_size + options.params.max_snail_size * limit
        )
        samples.append((i + 1, survey_id, point.longitude, point.latitude, seq, size))

    # Convert to dataframe.
    df = pl.DataFrame(samples, schema=('sample_id', 'survey_id', 'lon', 'lat', 'sequence', 'size'), orient='row')
    return df.with_columns(
        lon=df['lon'].round(LON_LAT_PRECISION),
        lat=df['lat'].round(LON_LAT_PRECISION),
        size=df['size'].round(SIZE_PRECISION),
    )


def _load_grids(options):
    '''Load all grid files.'''
    return {
        s: np.loadtxt(Path(options.grids, f'{s}.csv'), dtype=int, delimiter=',')
        for s in set(options.surveys['site_id'])
    }


def _random_geo(sites, surveys, grids):
    '''Select random (lon, lat) point from a randomly-selected sample grid.'''
    # Get parameters.
    survey_row = random.randrange(surveys.shape[0])
    survey_id = surveys.item(survey_row, 'survey_id')
    spacing = float(surveys.item(survey_row, 'spacing'))
    site_id = surveys.item(survey_row, 'site_id')
    site_row = sites['site_id'].to_list().index(site_id)
    site_lon = sites.item(site_row, 'lon')
    site_lat = sites.item(site_row, 'lat')

    # Get grid information.
    grid = grids[site_id]
    width, height = grid.shape
    rand_x, rand_y = random.randrange(width), random.randrange(height)
    contaminated = bool(grid[rand_x, rand_x] != 0)

    # Generate point.
    corner = lonlat(site_lon, site_lat)
    rand_x *= spacing
    rand_y *= spacing
    dist = math.sqrt(rand_x**2 + rand_y**2)
    bearing = math.degrees(math.atan2(rand_y, rand_x))
    point = distance(meters=dist).destination(corner, bearing=bearing)

    return survey_id, point, contaminated


def _save(options, samples):
    '''Save or show results.'''
    if options.outfile:
        samples.write_csv(Path(options.outfile))
    else:
        samples.write_csv(sys.stdout)


def _size_limit(options, genomes, seq, contaminated):
    '''Calculate upper bound on snail size.'''
    susc_loc = genomes['susceptible_loc']
    susc_base = genomes['susceptible_base']
    if contaminated and (seq[susc_loc] == susc_base):
        return options.params.mutant
    return options.params.normal



'''Mangle readings files.'''

import csv
from pathlib import Path
import random
import sqlite3


SELECT = '''
select
    plate.plate_id as plate_id,
    plate.filename as filename,
    plate.date as date,
    staff.staff_id as staff_id,
    staff.personal as personal,
    staff.family as family
from
    plate join performed join staff
on
    (plate.sample_id = performed.sample_id)
    and
    (performed.staff_id = staff.staff_id)
'''


def mangle(options):
    '''Main driver.'''
    con = sqlite3.connect(options.dbfile)
    con.row_factory = sqlite3.Row
    rows = list(dict(r) for r in con.execute(SELECT).fetchall())
    random.seed(len(rows))
    rows = _consolidate(rows)
    for row in rows:
        _mangle_file(options, row)


def _consolidate(rows):
    '''Pick a single record at random for each physical plate.'''
    grouped = {}
    for r in rows:
        if r['plate_id'] not in grouped:
            grouped[r['plate_id']] = []
        grouped[r['plate_id']].append(r)

    result = []
    for group in grouped.values():
        result.append(random.choice(group))
    return result


def _mangle_file(options, settings):
    '''Mangle a single file.'''
    sections = _read_sections(options, settings['filename'])
    for func in (_do_staff_name, _do_date, _do_footer, _do_indent,):
        if random.random() < func.prob:
            func(settings, sections)
    _write_sections(options, settings['filename'], sections)


def _do_date(settings, sections):
    '''Mangle by adding date in header.'''
    row = [''] * len(sections['header'][0])
    row[0] = 'Date'
    row[1] = settings['date']
    sections['header'].append(row)
_do_date.prob = 0.1


def _do_footer(settings, sections):
    '''Mangle by adding a footer.'''
    blank = [''] * len(sections['header'][0])
    foot = [''] * len(sections['header'][0])
    foot[0] = settings['staff_id']
    sections['footer'] = [blank, foot]
_do_footer.prob = 0.1


def _do_indent(settings, sections):
    '''Mangle by indenting.'''
    for section in sections.values():
        for row in section:
            row.insert(0, '')
_do_indent.prob = 0.1


def _do_staff_name(settings, sections):
    '''Mangle by adding staff name.'''
    sections['header'][0][-2] = f'{settings["personal"]} {settings["family"]}'
_do_staff_name.prob = 0.1


def _read_sections(options, filename):
    '''Read tidy file and split into sections.'''
    with open(Path(options.tidy, filename), 'r') as raw:
        rows = [row for row in csv.reader(raw)]
    return {
        'header': rows[0:1],
        'headspace': rows[1:2],
        'body': rows[2:],
        'footer': []
    }


def _write_sections(options, filename, sections):
    '''Write sections of mangled file.'''
    with open(Path(options.outdir, filename), 'w') as raw:
        writer = csv.writer(raw, lineterminator='\n')
        for section in sections.values():
            writer.writerows(section)



'''Generate database from data files.'''

import json
import polars as pl


ISO_DATE = "%Y-%m-%d"


def db(options):
    '''Main driver.'''
    url = f'sqlite:///{options.dbfile}'

    _csv_to_db(url, 'staff', options.staff)
    _csv_to_db(url, 'sample', options.samples)
    _csv_to_db(url, 'site', options.sites)
    _csv_to_db(url, 'survey', options.surveys, dates=['date'], columns=['survey_id', 'site_id', 'date'])

    assays = json.load(open(options.assays, 'r'))
    _json_to_db(url, assays, 'experiment', dates=['start', 'end'])
    _json_to_db(url, assays, 'performed')
    _json_to_db(url, assays, 'plate', dates=['date'])
    _json_to_db(url, assays, 'invalidated', dates=['date'])


def _csv_to_db(url, name, source, dates=[], columns=[]):
    '''Create table from CSV.'''
    df = pl.read_csv(source)
    if columns:
        df = df[columns]
    df = _convert_dates(df, dates)
    df.write_database(name, url, if_table_exists='replace')


def _json_to_db(url, data, name, dates=[]):
    '''Create table from JSON.'''
    df = pl.DataFrame(data[name])
    df = _convert_dates(df, dates)
    df.write_database(name, url, if_table_exists='replace')


def _convert_dates(df, dates):
    '''Convert text columns to dates.'''
    for colname in dates:
        df = df.with_columns(pl.col(colname).str.to_date(ISO_DATE))
    return df
